# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import argparse
import importlib
import logging
import os
import sys
import traceback
from collections.abc import Callable

try:
    import argcomplete
except ImportError:
    argcomplete = None

LOG = logging.getLogger(__name__)


class Command:
    """Base class for commands."""

    accept_unknown_args = False

    def run(self, args: argparse.Namespace) -> None:
        """Run the command.

        Parameters
        ----------
        args : argparse.Namespace
            The arguments for the command
        """
        raise NotImplementedError(f"Command not implemented: {args.command}")


def make_parser(description: str, commands: dict[str, Command]) -> argparse.ArgumentParser:
    """Create an argument parser for the CLI.

    Parameters
    ----------
    description : str
        The description of the CLI
    commands : dict[str, Command]
        A dictionary of command names to Command instances

    Returns
    -------
    argparse.ArgumentParser
        The argument parser
    """
    parser = argparse.ArgumentParser(
        description=description,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument(
        "--version",
        "-V",
        action="store_true",
        help="show the version and exit",
    )
    parser.add_argument(
        "--debug",
        "-d",
        action="store_true",
        help="Debug mode",
    )

    subparsers = parser.add_subparsers(help="commands:", dest="command")
    for name, command in commands.items():
        command_parser = subparsers.add_parser(name, description=command.__doc__, help=command.__doc__)
        command.add_arguments(command_parser)

    return parser


class Failed(Command):
    """Command not available."""

    def __init__(self, name: str, error: ImportError):
        """Initialize the Failed command.

        Parameters
        ----------
        name : str
            The name of the command
        error : ImportError
            The error that occurred
        """
        self.name = name
        self.error = error
        traceback.print_tb(error.__traceback__)

    def add_arguments(self, command_parser: argparse.ArgumentParser) -> None:
        """Add arguments to the command parser.

        Parameters
        ----------
        command_parser : argparse.ArgumentParser
            The command parser
        """
        command_parser.add_argument("x", nargs=argparse.REMAINDER)

    def run(self, args: argparse.Namespace) -> None:
        """Run the command.

        Parameters
        ----------
        args : argparse.Namespace
            The arguments for the command
        """
        print(f"Command '{self.name}' not available: {self.error}")
        sys.exit(1)


def register_commands(here: str, package: str, select: Callable, fail: Callable = None) -> dict[str, Command]:
    """Register commands from a package.

    Parameters
    ----------
    here : str
        The directory containing the commands
    package : str
        The package name
    select : Callable
        A function to select the command object from the module
    fail : Callable, optional
        A function to create a Failed command if a command cannot be imported

    Returns
    -------
    dict[str, Command]
        A dictionary of command names to Command instances
    """
    result = {}
    not_available = {}

    for p in sorted(os.listdir(here)):
        full = os.path.join(here, p)
        if p.startswith("_"):
            continue
        if not (p.endswith(".py") or (os.path.isdir(full) and os.path.exists(os.path.join(full, "__init__.py")))):
            continue

        name, _ = os.path.splitext(p)

        try:
            imported = importlib.import_module(
                f".{name}",
                package=package,
            )
        except ImportError as e:
            not_available[name] = e
            continue

        obj = select(imported)
        if obj is None:
            continue

        if hasattr(obj, "command"):
            name = obj.command

        if name in result:
            msg = f"Duplicate command '{name}', please choose a different command name for {type(obj)}"
            raise ValueError(msg)
        if " " in name:
            msg = f"Commands cannot contain spaces: '{name}' in {type(obj)}"
            raise ValueError(msg)

        result[name] = obj

    for name, e in not_available.items():
        if fail is None:
            pass
        if callable(fail):
            result[name] = fail(name, e)

    return result


def cli_main(
    version: str, description: str, commands: dict[str, Command], test_arguments: list[str] | None = None
) -> None:
    """Main entry point for the CLI.

    Parameters
    ----------
    version : str
        The version of the CLI
    description : str
        The description of the CLI
    commands : dict[str, Command]
        A dictionary of command names to Command instances
    test_arguments : list[str], optional
        The command line arguments to parse, used for testing purposes, by default None
    """
    parser = make_parser(description, commands)
    args, unknown = parser.parse_known_args(test_arguments)
    if argcomplete:
        argcomplete.autocomplete(parser)

    if args.version:
        print(version)
        return

    if args.command is None:
        parser.print_help()
        return

    cmd = commands[args.command]

    logging.basicConfig(
        format="%(asctime)s %(levelname)s %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        level=logging.DEBUG if args.debug else logging.INFO,
    )

    if unknown and not cmd.accept_unknown_args:
        # This should trigger an error
        parser.parse_args(test_arguments)

    try:
        if unknown:
            cmd.run(args, unknown)
        else:
            cmd.run(args)
    except ValueError as e:

        if test_arguments:
            raise

        traceback.print_exc()
        LOG.error("\n💣 %s", str(e).lstrip())
        LOG.error("💣 Exiting")
        sys.exit(1)

    if not test_arguments:
        sys.exit(0)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from __future__ import annotations

from typing import Any

try:
    from mlflow import MlflowClient
except ImportError:
    raise ImportError(
        "The `mlflow` package is required to use AnemoiMLflowclient. Please install it with `pip install mlflow`."
    )

from .auth import TokenAuth
from .utils import health_check


class AnemoiMlflowClient(MlflowClient):
    """Anemoi extension of the MLflow client with token authentication support."""

    def __init__(
        self,
        tracking_uri: str,
        *args,
        authentication: bool = False,
        check_health: bool = True,
        **kwargs,
    ) -> None:
        """Behaves like a normal `mlflow.MlflowClient` but with token authentication injected on every call.

        Parameters
        ----------
        tracking_uri : str
            The URI of the MLflow tracking server.
        authentication : bool, optional
            Enable token authentication, by default False
        check_health : bool, optional
            Check the health of the MLflow server on init, by default True
        *args : Any
            Additional arguments to pass to the MLflow client.
        **kwargs : Any
            Additional keyword arguments to pass to the MLflow client.

        """
        self.anemoi_auth = TokenAuth(tracking_uri, enabled=authentication)
        if check_health:
            super().__getattribute__("anemoi_auth").authenticate()
            health_check(tracking_uri)
        super().__init__(tracking_uri, *args, **kwargs)

    def __getattribute__(self, name: str) -> Any:
        """Intercept attribute access and inject authentication."""
        attr = super().__getattribute__(name)
        if callable(attr) and name != "anemoi_auth":
            super().__getattribute__("anemoi_auth").authenticate()
        return attr

    def login(self, force_credentials: bool = False, **kwargs) -> None:
        """Explicitly log in to the MLflow server by acquiring or refreshing the token.

        Parameters
        ----------
        force_credentials : bool, optional
            Force a credential login even if a refresh token is available, by default False.
        kwargs : dict
            Additional keyword arguments passed to the underlying TokenAuth.login.
        """
        self.anemoi_auth.login(force_credentials=force_credentials, **kwargs)



# (C) Copyright 2025- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
import os
import shutil
import warnings
from collections.abc import Callable
from functools import lru_cache
from pathlib import Path

import pytest
from multiurl import download

from anemoi.utils.humanize import list_to_human

LOG = logging.getLogger(__name__)

TEST_DATA_URL = "https://object-store.os-api.cci1.ecmwf.int/ml-tests/test-data/samples/"


def _check_path(path: str) -> None:
    """Check if the given path is normalized, not absolute, and does not start with a dot.

    Parameters
    ----------
    path : str
        The path to check.

    Raises
    ------
    AssertionError
        If the path is not normalized, is absolute, or starts with a dot.
    """
    assert os.path.normpath(path) == path, f"Path '{path}' should be normalized"
    assert not os.path.isabs(path), f"Path '{path}' should not be absolute"
    assert not path.startswith("."), f"Path '{path}' should not start with '.'"


class TemporaryDirectoryForTestData:
    def __init__(self, base_dir: Path) -> None:
        self.base_dir = base_dir

    def __call__(self, path: str = "", archive: bool = False) -> str:
        if path == "":
            return str(self.base_dir)
        _check_path(path)
        return str(self.base_dir.joinpath(*Path(path).parts)) + (".extracted" if archive else "")


@pytest.fixture(scope="session")
def temporary_directory_for_test_data(tmp_path_factory: pytest.TempPathFactory) -> TemporaryDirectoryForTestData:
    base_dir = tmp_path_factory.mktemp("test_data_base")
    return TemporaryDirectoryForTestData(base_dir)


def url_for_test_data(path: str) -> str:
    """Generate the URL for the test data based on the given path.

    Parameters
    ----------
    path : str
        The relative path to the test data.

    Returns
    -------
    str
        The full URL to the test data.
    """
    _check_path(path)

    return f"{TEST_DATA_URL}{path}"


class GetTestData:
    def __init__(self, temporary_directory_for_test_data: TemporaryDirectoryForTestData) -> None:
        self.temporary_directory_for_test_data = temporary_directory_for_test_data

    def __call__(self, path: str, gzipped: bool = False) -> str:
        """Download the test data to a temporary directory and return the local path.
        Parameters
        ----------
        path : str
            The relative path to the test data.
        gzipped : bool, optional
            Flag indicating if the remote file is gzipped, by default False. The local file will be gunzipped.

        Returns
        -------
        str
            The local path to the downloaded test data.
        """

        if _offline():
            raise RuntimeError("Offline mode: cannot download test data, add @pytest.mark.skipif(not offline(),...)")

        target = self.temporary_directory_for_test_data(path)

        if os.path.exists(target):
            return target

        os.makedirs(os.path.dirname(target), exist_ok=True)
        url = url_for_test_data(path)

        if gzipped:
            url += ".gz"
            target += ".gz"

        LOG.info(f"Downloading test data from {url} to {target}")

        download(url, target)

        if gzipped:
            import gzip

            with gzip.open(target, "rb") as f_in:
                with open(target[:-3], "wb") as f_out:
                    shutil.copyfileobj(f_in, f_out)
            os.remove(target)
            target = target[:-3]

        return target


@pytest.fixture()
def get_test_data(temporary_directory_for_test_data: TemporaryDirectoryForTestData) -> GetTestData:
    return GetTestData(temporary_directory_for_test_data)


class GetTestArchive:
    def __init__(
        self, temporary_directory_for_test_data: TemporaryDirectoryForTestData, get_test_data: GetTestData
    ) -> None:
        self.temporary_directory_for_test_data = temporary_directory_for_test_data
        self.get_test_data = get_test_data

    def __call__(self, path: str) -> Path:
        """Download an archive file (.zip, .tar, .tar.gz, .tar.bz2, .tar.xz) to a temporary directory
        unpack it, and return the local path to the directory containing the extracted files.

        Parameters
        ----------
        path : str
            The relative path to the test data.

        Returns
        -------
        str
            The local path to the downloaded test data.
        """

        target = Path(self.temporary_directory_for_test_data(path, archive=True))

        if os.path.exists(target):
            return target

        archive = self.get_test_data(path)

        shutil.unpack_archive(archive, os.path.dirname(target) + ".tmp")
        os.rename(os.path.dirname(target) + ".tmp", target)

        os.remove(archive)

        return target


@pytest.fixture()
def get_test_archive(
    temporary_directory_for_test_data: TemporaryDirectoryForTestData, get_test_data: GetTestData
) -> GetTestArchive:
    return GetTestArchive(temporary_directory_for_test_data, get_test_data)


def packages_installed(*names: str) -> bool:
    """Check if all the given packages are installed.

    Use this function to check if the required packages are installed before running tests.

    >>> @pytest.mark.skipif(not packages_installed("foo", "bar"), reason="Packages 'foo' and 'bar' are not installed")
    >>> def test_foo_bar() -> None:
    >>>    ...

    Parameters
    ----------
    names : str
        The names of the packages to check.

    Returns
    -------
    bool:
        Flag indicating if all the packages are installed."
    """

    warnings.warn(
        "The 'packages_installed' function is deprecated. Use '@skip_if_missing' instead.",
        DeprecationWarning,
        stacklevel=2,
    )

    for name in names:
        try:
            __import__(name)
        except ImportError:
            return False
    return True


def _missing_packages(*names: str) -> list[str]:
    """Check if the given packages are missing.

    Use this function to check if the required packages are missing before running tests.

    >>> @pytest.mark.skipif(missing_packages("foo", "bar"), reason="Packages 'foo' and 'bar' are not installed")
    >>> def test_foo_bar() -> None:
    >>>    ...

    Parameters
    ----------
    names : str
        The names of the packages to check.

    Returns
    -------
    list[str]:
        List of missing packages.
    """

    missing = []
    for name in names:
        try:
            __import__(name)
        except ImportError:
            missing.append(name)
    return missing


def _run_slow_tests() -> bool:
    """Check if the SLOW_TESTS environment variable is set.

    Returns
    -------
    bool
        True if the SLOW_TESTS environment variable is set, False otherwise.
    """
    return bool(int(os.environ.get("SLOW_TESTS", 0)))


@lru_cache(maxsize=None)
def _offline() -> bool:
    """Check if we are offline."""
    from urllib import request
    from urllib.error import URLError

    try:
        request.urlopen("https://anemoi.ecmwf.int", timeout=1)
        return False
    except URLError:
        return True


skip_if_offline = pytest.mark.skipif(_offline(), reason="No internet connection")
skip_slow_tests = pytest.mark.skipif(not _run_slow_tests(), reason="Skipping slow tests")


def skip_missing_packages(*names: str) -> pytest.MarkDecorator:
    """Skip a test if any of the specified packages are missing.

    Parameters
    ----------
    names : str
        The names of the packages to check.

    Returns
    -------
    Callable
        A decorator that skips the test if any of the specified packages are missing.
    """

    missing = [f"'{p}'" for p in _missing_packages(*names)]

    reason = ""

    if len(missing) == 1:
        reason = f"Package {missing[0]} is not installed"
    elif len(missing):
        reason = f"Packages {list_to_human(missing)} are not installed"

    return pytest.mark.skipif(len(missing) > 0, reason=reason)


def skip_if_missing_command(cmd: str) -> pytest.MarkDecorator:
    """Skip a test if the specified command is not available.

    Parameters
    ----------
    cmd : str
        The name of the command to check.

    Returns
    -------
    Callable
        A decorator that skips the test if the specified command is not available.
    """

    import shutil

    return pytest.mark.skipif(not shutil.which(cmd), reason=f"Command '{cmd}' is not available")


def cli_testing(package: str, cmd: str, *args: str) -> None:
    """Run a CLI command for testing purposes.

    Parameters
    ----------
    package : str
        The name of the package containing the CLI commands.
        Can be 'anemoi-datasets' or 'anemoi.datasets'.
    cmd : str
        The command to run.
    *args : str
        Additional arguments to pass to the command.
    """

    package = package.replace("-", ".")
    COMMANDS = getattr(__import__(f"{package}.commands", fromlist=["COMMANDS"]), "COMMANDS")
    version = getattr(__import__(f"{package}._version", fromlist=["__version__"]), "__version__", "0.1.0")

    from anemoi.utils.cli import cli_main

    cli_main(
        version=version,
        description=f"Testing the '{cmd}' CLI command from the '{package}' package.",
        commands=COMMANDS,
        test_arguments=[cmd] + list(args),
    )


def run_tests(globals: dict[str, Callable[[], None]]) -> None:
    """Run all test functions that start with 'test_'.

    Parameters
    ----------
    globals : dict[str, Callable[[], None]]
        The global namespace containing the test functions.

    Example
    -------

    Call from a test file to run all tests in that file:

    ```python
    if __name__ == "__main__":
        from anemoi.utils.testing import run_tests
        run_tests(globals())
    ```

    Useful for debugging or running tests in an interactive environment.

    """
    import logging

    import rich

    logging.basicConfig(level=logging.INFO)

    for name, obj in list(globals.items()):
        if name.startswith("test_") and callable(obj):
            pytestmark = getattr(obj, "pytestmark", None)
            if pytestmark is not None:
                if not isinstance(pytestmark, list):
                    pytestmark = [pytestmark]

                skip = False
                for m in pytestmark:
                    if m.name == "skipif" and m.args == (True,):
                        skip = True
                        rich.print(
                            f"[red]Skipping [bold]{name}[/bold] due to skipif condition [bold]{m.kwargs['reason']}[/bold].[/red]"
                        )
                        break
                if skip:
                    continue

            rich.print(f"[green]Running [bold]{name}[/bold]...[/green]")
            obj()



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import hashlib
import json
import os
import time
from collections.abc import Callable
from threading import Lock
from typing import Any

import numpy as np

LOCK = Lock()
CACHE = {}


def _get_cache_path(collection: str) -> str:
    """Get the cache path for a collection.

    Parameters
    ----------
    collection : str
        The name of the collection

    Returns
    -------
    str
        The cache path
    """
    return os.path.join(os.path.expanduser("~"), ".cache", "anemoi", collection)


def clean_cache(collection: str = "default") -> None:
    """Clean the cache for a collection.

    Parameters
    ----------
    collection : str, optional
        The name of the collection, by default "default"
    """
    global CACHE
    CACHE = {}
    path = _get_cache_path(collection)
    if not os.path.exists(path):
        return
    for filename in os.listdir(path):
        os.remove(os.path.join(path, filename))


class Cacher:
    """This class implements a simple caching mechanism.
    Private class, do not use directly.
    """

    def __init__(self, collection: str, expires: int | None):
        """Initialize the Cacher.

        Parameters
        ----------
        collection : str
            The name of the collection
        expires : int, optional
            The expiration time in seconds, or None for no expiration
        """
        self.collection = collection
        self.expires = expires

    def __call__(self, func: Callable) -> Callable:
        """Wrap a function with caching.

        Parameters
        ----------
        func : Callable
            The function to wrap

        Returns
        -------
        Callable
            The wrapped function
        """
        full = f"{func.__module__}.{func.__name__}"

        def wrapped(*args, **kwargs):
            with LOCK:
                return self.cache(
                    (full, args, kwargs),
                    lambda: func(*args, **kwargs),
                )

        return wrapped

    def cache(self, key: tuple, proc: Callable) -> Any:
        """Cache the result of a function.

        Parameters
        ----------
        key : tuple
            The cache key
        proc : Callable
            The function to call if the result is not cached

        Returns
        -------
        Any
            The cached result
        """
        key = json.dumps(key, sort_keys=True)
        m = hashlib.md5()
        m.update(key.encode("utf-8"))
        m = m.hexdigest()

        if m in CACHE:
            return CACHE[m]

        path = _get_cache_path(self.collection)

        filename = os.path.join(path, m) + self.ext
        if os.path.exists(filename):
            data = self.load(filename)
            if self.expires is None or data["expires"] > time.time():
                if data["key"] == key:
                    return data["value"]

        value = proc()
        data = {"key": key, "value": value}
        if self.expires is not None:
            data["expires"] = time.time() + self.expires

        os.makedirs(path, exist_ok=True)
        temp_filename = self.save(filename, data)
        os.rename(temp_filename, filename)

        CACHE[m] = value
        return value


class JsonCacher(Cacher):
    """Cacher that uses JSON files."""

    ext = ""

    def save(self, path: str, data: dict) -> str:
        """Save data to a JSON file.

        Parameters
        ----------
        path : str
            The path to the JSON file
        data : dict
            The data to save

        Returns
        -------
        str
            The temporary file path
        """
        temp_path = path + ".tmp"
        with open(temp_path, "w") as f:
            json.dump(data, f)
        return temp_path

    def load(self, path: str) -> dict:
        """Load data from a JSON file.

        Parameters
        ----------
        path : str
            The path to the JSON file

        Returns
        -------
        dict
            The loaded data
        """
        with open(path) as f:
            return json.load(f)


class NpzCacher(Cacher):
    """Cacher that uses NPZ files."""

    ext = ".npz"

    def save(self, path: str, data: dict) -> str:
        """Save data to an NPZ file.

        Parameters
        ----------
        path : str
            The path to the NPZ file
        data : dict
            The data to save

        Returns
        -------
        str
            The temporary file path
        """
        temp_path = path + ".tmp.npz"
        np.savez(temp_path, **data)
        return temp_path

    def load(self, path: str) -> dict:
        """Load data from an NPZ file.

        Parameters
        ----------
        path : str
            The path to the NPZ file

        Returns
        -------
        dict
            The loaded data
        """
        return np.load(path, allow_pickle=True)


# This function is the main entry point for the caching mechanism for the other anemoi packages
def cached(collection: str = "default", expires: int | None = None, encoding: str = "json") -> Callable:
    """Decorator to cache the result of a function.

    Default is to use a json file to store the cache, but you can also use npz files
    to cache dict of numpy arrays.

    Parameters
    ----------
    collection : str, optional
        The name of the collection, by default "default"
    expires : int, optional
        The expiration time in seconds, or None for no expiration, by default None
    encoding : str, optional
        The encoding type, either "json" or "npz", by default "json"

    Returns
    -------
    Callable
        The decorated function
    """
    return dict(json=JsonCacher, npz=NpzCacher)[encoding](collection, expires)



# (C) Copyright 2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


"""Utilities for working with grids."""

import logging
import os
from io import BytesIO

import deprecation
import numpy as np
import requests

from anemoi.utils._version import __version__

from .caching import cached

LOG = logging.getLogger(__name__)


GRIDS_URL_PATTERN = "https://get.ecmwf.int/repository/anemoi/grids/grid-{name}.npz"


@deprecation.deprecated(
    deprecated_in="0.4.25",
    removed_in="0.5.0",
    current_version=__version__,
    details="Use anemoi.transform.spatial.xyz_to_latlon instead.",
)
def xyz_to_latlon(x: np.ndarray, y: np.ndarray, z: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    """Convert Cartesian coordinates to latitude and longitude.

    Parameters
    ----------
    x : np.ndarray
        The x coordinates
    y : np.ndarray
        The y coordinates
    z : np.ndarray
        The z coordinates

    Returns
    -------
    tuple[np.ndarray, np.ndarray]
        The latitude and longitude
    """
    return (
        np.rad2deg(np.arcsin(np.minimum(1.0, np.maximum(-1.0, z)))),
        np.rad2deg(np.arctan2(y, x)),
    )


@deprecation.deprecated(
    deprecated_in="0.4.25",
    removed_in="0.5.0",
    current_version=__version__,
    details="Use anemoi.transform.spatial.xyz_to_latlon instead.",
)
def latlon_to_xyz(lat: np.ndarray, lon: np.ndarray, radius: float = 1.0) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Convert latitude and longitude to Cartesian coordinates.

    Parameters
    ----------
    lat : np.ndarray
        The latitudes
    lon : np.ndarray
        The longitudes
    radius : float, optional
        The radius of the sphere, by default 1.0

    Returns
    -------
    tuple[np.ndarray, np.ndarray, np.ndarray]
        The x, y, and z coordinates
    """
    # https://en.wikipedia.org/wiki/Geographic_coordinate_conversion#From_geodetic_to_ECEF_coordinates
    # We assume that the Earth is a sphere of radius 1 so N(phi) = 1
    # We assume h = 0
    #
    phi = np.deg2rad(lat)
    lda = np.deg2rad(lon)

    cos_phi = np.cos(phi)
    cos_lda = np.cos(lda)
    sin_phi = np.sin(phi)
    sin_lda = np.sin(lda)

    x = cos_phi * cos_lda * radius
    y = cos_phi * sin_lda * radius
    z = sin_phi * radius

    return x, y, z


@deprecation.deprecated(
    deprecated_in="0.4.25",
    removed_in="0.5.0",
    current_version=__version__,
    details="Use anemoi.transform.spatial.nearest_grid_points instead.",
)
def nearest_grid_points(
    source_latitudes: np.ndarray,
    source_longitudes: np.ndarray,
    target_latitudes: np.ndarray,
    target_longitudes: np.ndarray,
) -> np.ndarray:
    """Find the nearest grid points.

    Parameters
    ----------
    source_latitudes : np.ndarray
        The source latitudes
    source_longitudes : np.ndarray
        The source longitudes
    target_latitudes : np.ndarray
        The target latitudes
    target_longitudes : np.ndarray
        The target longitudes

    Returns
    -------
    np.ndarray
        The indices of the nearest grid points
    """
    from scipy.spatial import cKDTree

    source_xyz = latlon_to_xyz(source_latitudes, source_longitudes)
    source_points = np.array(source_xyz).transpose()

    target_xyz = latlon_to_xyz(target_latitudes, target_longitudes)
    target_points = np.array(target_xyz).transpose()

    _, indices = cKDTree(source_points).query(target_points, k=1)
    return indices


@cached(collection="grids", encoding="npz")
def _grids(name: str | list[float] | tuple[float, ...]) -> bytes:
    """Get grid data by name.

    Parameters
    ----------
    name : str
        The name of the grid

    Returns
    -------
    bytes
        The grid data
    """
    from anemoi.utils.config import load_config

    if isinstance(name, (tuple, list)):
        assert len(name) == 2, "Grid name must be a list or a tuple of length 2"
        assert all(isinstance(i, (int, float)) for i in name), "Grid name must be a list or a tuple of numbers"
        if name[0] == name[1]:
            name = str(float(name[0]))
        else:
            name = str(float(name[0])) + "x" + str(float(name[1]))
        name = name.replace(".", "p")

    user_path = load_config().get("utils", {}).get("grids_path")
    if user_path:
        path = os.path.expanduser(os.path.join(user_path, f"grid-{name}.npz"))
        if os.path.exists(path):
            LOG.warning("Loading grids from custom user path %s", path)
            with open(path, "rb") as f:
                return f.read()
        else:
            LOG.warning("Custom user path %s does not exist", path)

    # To add a grid
    # anemoi-transform get-grid --source mars grid=o400,levtype=sfc,param=2t grid-o400.npz
    # nexus-cli -u xxxx -p yyyy -s GET_INSTANCE --repository anemoi upload --remote-path grids --local-path grid-o400.npz

    url = GRIDS_URL_PATTERN.format(name=name.lower())
    LOG.warning("Downloading grids from %s", url)
    response = requests.get(url)
    response.raise_for_status()
    return response.content


@deprecation.deprecated(
    deprecated_in="0.4.25",
    removed_in="0.5.0",
    current_version=__version__,
    details="Use anemoi.transform.grids.named.lookup instead.",
)
def grids(name: str | list[float] | tuple[float, ...]) -> dict:
    """Load grid data by name.

    Parameters
    ----------
    name : str
        The name of the grid

    Returns
    -------
    dict
        The grid data
    """
    if isinstance(name, str) and name.endswith(".npz"):
        return dict(np.load(name))

    data = _grids(name)
    npz = np.load(BytesIO(data))
    return dict(npz)



# (C) Copyright 2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from collections.abc import Mapping
from typing import Any


class Rule:

    def __init__(self, match: dict[str, Any], result: Any):
        """Initialize a Rule object.

        Parameters
        ----------
        match : Dict[str, Any]
            A dictionary defining the conditions for the rule to match.
        result : Any
            The result to return if the rule matches.
        """
        self._match = match
        self._result = result

    def match(self, obj: Mapping[str, Any]) -> bool:
        """Check if the rule matches the given object.

        Parameters
        ----------
        obj : Mapping[str, Any]
            The object to check against the rule's conditions.

        Returns
        -------
        bool
            True if the rule matches, False otherwise.
        """
        for key, value in self._match.items():
            if key not in obj or obj[key] != value:
                return False

        return True

    @property
    def result(self) -> Any:
        """The result associated with the rule."""
        return self._result

    @property
    def condition(self) -> dict[str, Any]:
        """The conditions that define the rule."""
        return self._match


class RuleSet:

    def __init__(self, rules: list[Rule | dict[str, Any] | list[Any]]):
        """Initialize a RuleSet object.

        Parameters
        ----------
        rules : List[Union[Rule, Dict[str, Any], List[Any]]]
            A list of rules, where each rule can be a Rule object, a dictionary with
            'match' and 'result' keys, or a list with two elements (match and result).
        """
        assert isinstance(rules, list), "rules must be a list"

        self.rules: list[Rule] = []

        for rule in rules:
            if isinstance(rule, Rule):
                self.rules.append(rule)
                continue

            if isinstance(rule, dict):

                assert len(rule) == 2, "Rule dictionary must contain exactly two key-value pair."

                match = rule.get("match")
                if match is None:
                    raise ValueError("Rule dictionary must contain a 'match' key.")

                result = rule.get("result")
                if result is None:
                    raise ValueError("Rule dictionary must contain a 'result' key.")

                self.rules.append(Rule(match, result))
                continue

            if isinstance(rule, list):
                assert len(rule) == 2, "Rule list must contain exactly two elements."
                match = rule[0]
                result = rule[1]
                self.rules.append(Rule(match, result))
                continue

            raise ValueError(
                "Rule must be either a Rule object, a dictionary with 'match' and 'result' keys, or a list with two elements."
            )

    @classmethod
    def from_list(cls, rules: list[Any]) -> "RuleSet":
        """Create a RuleSet from a list of rules.

        Parameters
        ----------
        rules : List[Any]
            A list of rules to initialize the RuleSet.

        Returns
        -------
        RuleSet
            A new RuleSet object.
        """
        return cls(rules)

    @classmethod
    def from_files(cls, path: str) -> "RuleSet":
        """Create a RuleSet from a file.

        Parameters
        ----------
        path : str
            The path to the file containing the rules. Supported formats are .json and .yaml/.yml.

        Returns
        -------
        RuleSet
            A new RuleSet object.

        Raises
        ------
        ValueError
            If the file format is unsupported.
        """
        if path.endswith(".json"):
            import json

            with open(path) as f:
                return cls.from_list(json.load(f))

        if path.endswith(".yaml") or path.endswith(".yml"):
            import yaml

            with open(path) as f:
                return cls.from_list(yaml.safe_load(f))

        raise ValueError("Unsupported file format. Supported formats are .json and .yaml/.yml.")

    @classmethod
    def from_any(cls, rules: str | list[Any]) -> "RuleSet":
        """Create a RuleSet from a list or a file path.

        Parameters
        ----------
        rules : Union[str, List[Any]]
            The rules to initialize the RuleSet, either as a list or a file path.

        Returns
        -------
        RuleSet
            A new RuleSet object.

        Raises
        ------
        ValueError
            If the rules format is unsupported.
        """
        if isinstance(rules, str):
            return cls.from_files(rules)

        if isinstance(rules, list):
            return cls.from_list(rules)

        raise ValueError("Unsupported rules format. Must be a list or a file path.")

    def match(self, obj: Mapping[str, Any], strategy: str = "first-match") -> Rule | None:
        """Match an object against the rules in the RuleSet.

        Parameters
        ----------
        obj : Mapping[str, Any]
            The object to match against the rules.
        strategy : str, optional
            The matching strategy to use. Currently, only 'first-match' is supported.

        Returns
        -------
        Optional[Rule]
            The first matching rule, or None if no match is found.

        Raises
        ------
        AssertionError
            If an unsupported strategy is provided.
        """
        assert strategy == "first-match", "Only 'first-match' strategy is supported for now."
        for rule in self.rules:
            if rule.match(obj):
                return rule

        return None

    def __iter__(self) -> iter:
        """Return an iterator over the rules in the RuleSet.

        Returns
        -------
        iter
            An iterator over the Rule objects in the RuleSet.
        """
        return iter(self.rules)

    def __len__(self) -> int:
        """Return the number of rules in the RuleSet.

        Returns
        -------
        int
            The number of rules in the RuleSet.
        """
        return len(self.rules)

    def __getitem__(self, index: int) -> Rule:
        """Retrieve a rule by its index.

        Parameters
        ----------
        index : int
            The index of the rule to retrieve.

        Returns
        -------
        Rule
            The rule at the specified index.
        """
        return self.rules[index]



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime

from anemoi.utils.dates import frequency_to_string
from anemoi.utils.dates import frequency_to_timedelta


def test_frequency_to_string() -> None:
    """Test the frequency_to_string function for converting timedelta to string."""
    assert frequency_to_string(datetime.timedelta(hours=1)) == "1h"
    assert frequency_to_string(datetime.timedelta(hours=1, minutes=30)) == "1:30:00"
    assert frequency_to_string(datetime.timedelta(days=10)) == "10d"
    assert frequency_to_string(datetime.timedelta(minutes=10)) == "10m"
    assert frequency_to_string(datetime.timedelta(minutes=90)) == "1:30:00"


def test_frequency_to_timedelta() -> None:
    """Test the frequency_to_timedelta function for converting string to timedelta."""
    assert frequency_to_timedelta("1s") == datetime.timedelta(seconds=1)
    assert frequency_to_timedelta("3m") == datetime.timedelta(minutes=3)
    assert frequency_to_timedelta("1h") == datetime.timedelta(hours=1)
    assert frequency_to_timedelta("3d") == datetime.timedelta(days=3)
    assert frequency_to_timedelta("90m") == datetime.timedelta(hours=1, minutes=30)
    assert frequency_to_timedelta("0:30") == datetime.timedelta(minutes=30)
    assert frequency_to_timedelta("0:30:10") == datetime.timedelta(minutes=30, seconds=10)
    assert frequency_to_timedelta("1:30:10") == datetime.timedelta(hours=1, minutes=30, seconds=10)

    assert frequency_to_timedelta("PT10M") == datetime.timedelta(minutes=10)


if __name__ == "__main__":
    for name, obj in list(globals().items()):
        if name.startswith("test_") and callable(obj):
            print(f"Running {name}...")
            obj()



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    import pytest_mock

from anemoi.utils.mlflow.client import AnemoiMlflowClient


@pytest.fixture(autouse=True)
def mocks(mocker: pytest_mock.MockerFixture) -> None:
    mocker.patch("anemoi.utils.mlflow.client.TokenAuth")
    mocker.patch("anemoi.utils.mlflow.client.health_check")
    mocker.patch("anemoi.utils.mlflow.client.AnemoiMlflowClient.search_experiments")


def test_auth_injected() -> None:
    client = AnemoiMlflowClient("http://localhost:5000", authentication=True, check_health=False)
    client.search_experiments()
    client.search_experiments()

    assert client.anemoi_auth.authenticate.call_count == 2


def test_health_check() -> None:
    # the internal health check will trigger an authenticate call
    client = AnemoiMlflowClient("http://localhost:5000", authentication=True, check_health=True)

    client.anemoi_auth.authenticate.assert_called_once()


def test_login_delegates_to_token_auth() -> None:
    client = AnemoiMlflowClient("http://localhost:5000", authentication=True, check_health=False)
    client.login(force_credentials=True)
    client.anemoi_auth.login.assert_called_once_with(force_credentials=True)



# (C) Copyright 2025 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from pydantic import BaseModel as PydanticBaseModel


class BaseModel(PydanticBaseModel):
    class Config:
        """Pydantic BaseModel configuration."""

        use_attribute_docstrings = True
        use_enum_values = True
        validate_assignment = True
        validate_default = True
        extra = "forbid"



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import warnings
from collections.abc import Callable
from typing import Any

from .remote import transfer
from .remote.s3 import delete as delete_
from .remote.s3 import s3_client as s3_client_

warnings.warn(
    "The anemoi.utils.s3 module is deprecated and will be removed in a future release. "
    "Please use the 'anemoi.utils.remote' or 'anemoi.utils.remote.s3' module instead.",
    DeprecationWarning,
    stacklevel=2,
)


def s3_client(*args: Any, **kwargs: Any) -> Any:
    """Create an S3 client.

    Parameters
    ----------
    *args : Any
        Positional arguments for the S3 client.
    **kwargs : Any
        Keyword arguments for the S3 client.

    Returns
    -------
    Any
        The S3 client.
    """
    warnings.warn(
        "The 's3_client' function (from anemoi.utils.s3 import s3_client) function is deprecated and will be removed in a future release. "
        "Please use the 's3_client' function (from anemoi.utils.remote.s3 import s3_client) instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return s3_client_(*args, **kwargs)


def upload(
    source: str,
    target: str,
    *,
    overwrite: bool = False,
    resume: bool = False,
    verbosity: int = 1,
    progress: Callable | None = None,
    threads: int = 1,
) -> None:
    """Upload a file to S3.

    Parameters
    ----------
    source : str
        The source file path.
    target : str
        The target S3 path.
    overwrite : bool, optional
        Whether to overwrite the target file, by default False.
    resume : bool, optional
        Whether to resume a previous upload, by default False.
    verbosity : int, optional
        The verbosity level, by default 1.
    progress : Callable, optional
        A callback function for progress updates, by default None.
    threads : int, optional
        The number of threads to use, by default 1.
    """
    warnings.warn(
        "The 'upload' function (from anemoi.utils.s3 import upload) function is deprecated and will be removed in a future release. "
        "Please use the 'transfer' function (from anemoi.utils.remote import transfer) instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return transfer(
        source, target, overwrite=overwrite, resume=resume, verbosity=verbosity, progress=progress, threads=threads
    )


def download(*args: Any, **kwargs: Any) -> Any:
    """Download a file from S3.

    Parameters
    ----------
    *args : Any
        Positional arguments for the download.
    **kwargs : Any
        Keyword arguments for the download.

    Returns
    -------
    Any
        The result of the download.
    """
    warnings.warn(
        "The 'download' function (from anemoi.utils.s3 import download) function is deprecated and will be removed in a future release. "
        "Please use the 'transfer' function (from anemoi.utils.remote import transfer) instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return transfer(*args, **kwargs)


def delete(*args: Any, **kwargs: Any) -> Any:
    """Delete a file from S3.

    Parameters
    ----------
    *args : Any
        Positional arguments for the delete.
    **kwargs : Any
        Keyword arguments for the delete.

    Returns
    -------
    Any
        The result of the delete.
    """
    warnings.warn(
        "The 'delete' function (from anemoi.utils.s3 import delete) function is deprecated and will be removed in a future release. "
        "Please use the 'transfer' function (from anemoi.utils.remote.s3 import delete) instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return delete_(*args, **kwargs)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from __future__ import annotations

import functools
from collections.abc import Callable
from typing import Any


def aliases(aliases: dict[str, str | list[str]] | None = None, **kwargs: Any) -> Callable[[Callable], Callable]:
    """Alias keyword arguments in a function call.

    Allows for dynamically renaming keyword arguments in a function call.

    Parameters
    ----------
    aliases : dict[str, Union[str, list[str]]], optional
        Key, value pair of aliases, with keys being the true name, and value being a str or list of aliases,
        by default None
    **kwargs : Any
        Kwargs form of aliases

    Returns
    -------
    Callable
        Decorator function that renames keyword arguments in a function call.

    Raises
    ------
    ValueError
        If the aliasing would result in duplicate keys.

    Examples
    --------
    ```python
    @aliases(a="b", c=["d", "e"])
    def func(a, c):
        return a, c

    func(a=1, c=2)  # (1, 2)
    func(b=1, d=2)  # (1, 2)
    ```
    """

    if aliases is None:
        aliases = {}
    aliases.update(kwargs)

    aliases = {v: k for k, vs in aliases.items() for v in (vs if isinstance(vs, list) else [vs])}

    def decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            keys = kwargs.keys()
            for k in set(keys).intersection(set(aliases.keys())):
                if aliases[k] in keys:
                    raise ValueError(
                        f"When aliasing {k} with {aliases[k]} duplicate keys were present. Cannot include both."
                    )
                kwargs[aliases[k]] = kwargs.pop(k)

            return func(*args, **kwargs)

        return wrapper

    return decorator



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from __future__ import annotations

import os
import time

import pytest

from anemoi.utils.mlflow.auth import TokenAuth


def mocks(
    mocker: pytest.MockerFixture,
    token_request: dict | None = None,
    load_config: dict | None = None,
) -> pytest.Mock:
    if load_config is None:
        load_config = {}
    if token_request is None:
        token_request = {}
    response = {
        "access_token": "access_token",
        "expires_in": 3600,
        "refresh_token": "new_refresh_token",
    }
    response.update(token_request)

    config = {
        "refresh_token": "old_refresh_token",
        "refresh_expires": time.time() + 3600,
    }
    config.update(load_config)

    mock_token_request = mocker.patch(
        "anemoi.utils.mlflow.auth.TokenAuth._token_request",
        return_value=response,
    )
    mocker.patch(
        "anemoi.utils.mlflow.auth.load_config",
        return_value=config,
    )
    mocker.patch(
        "anemoi.utils.mlflow.auth.save_config",
    )
    mocker.patch(
        "anemoi.utils.mlflow.auth.getpass",
        return_value="seed_refresh_token",
    )
    mocker.patch("os.environ")

    return mock_token_request


def test_auth(mocker: pytest.MockerFixture) -> None:
    mock_token_request = mocks(mocker)

    auth = TokenAuth("https://test.url")

    assert auth.access_token is None
    assert auth.refresh_token == "old_refresh_token"  # noqa: S105

    auth.authenticate()
    # test that no new token is requested the second time
    auth.authenticate()

    mock_token_request.assert_called_once()

    assert auth.access_token == "access_token"  # noqa: S105
    assert auth.access_expires > time.time()
    assert auth.refresh_token == "new_refresh_token"  # noqa: S105


def test_not_logged_in(mocker: pytest.MockerFixture) -> None:
    # no refresh token
    mocks(mocker, load_config={"refresh_token": None})
    auth = TokenAuth("https://test.url")
    pytest.raises(RuntimeError, auth.authenticate)

    # expired refresh token
    mocks(mocker, load_config={"refresh_expires": time.time() - 1})
    auth = TokenAuth("https://test.url")
    pytest.raises(RuntimeError, auth.authenticate)


def test_login(mocker: pytest.MockerFixture) -> None:
    # normal login
    mock_token_request = mocks(mocker)
    auth = TokenAuth("https://test.url")
    auth.login()

    mock_token_request.assert_called_once()

    # normal credential login
    mock_token_request = mocks(mocker, load_config={"refresh_token": None})
    auth = TokenAuth("https://test.url")
    auth.login()

    mock_token_request.assert_called_once()

    # forced credential login
    mock_token_request = mocks(mocker)
    auth = TokenAuth("https://test.url")
    auth.login(force_credentials=True)

    mock_token_request.assert_called_once()

    # failed login
    mock_token_request = mocks(mocker, token_request={"refresh_token": None})
    auth = TokenAuth("https://test.url")
    pytest.raises(RuntimeError, auth.login)

    assert mock_token_request.call_count == 2


def test_enabled(mocker: pytest.MockerFixture) -> None:
    mock_token_request = mocks(mocker)
    auth = TokenAuth("https://test.url", enabled=False)
    auth.authenticate()

    mock_token_request.assert_not_called()


def test_api(mocker: pytest.MockerFixture) -> None:
    mocks(mocker)
    auth = TokenAuth("https://test.url")
    mock_post = mocker.patch("requests.post")

    # successful request
    response_json = {"status": "OK", "response": {}}
    mocker.patch("requests.post.return_value.json", return_value=response_json)
    response = auth._request("path", {"key": "value"})

    assert response == response_json["response"]
    mock_post.assert_called_once_with(
        "https://test.url/path",
        json={"key": "value"},
        headers=mocker.ANY,
        timeout=mocker.ANY,
    )

    # api error
    error_response = {"status": "ERROR", "response": {}}
    mocker.patch("requests.post.return_value.json", return_value=error_response)

    with pytest.raises(RuntimeError):
        auth._request("path", {"key": "value"})


def test_target_env_var(mocker: pytest.MockerFixture) -> None:
    mocks(mocker)
    auth = TokenAuth("https://test.url", target_env_var="MLFLOW_TEST_ENV_VAR")
    auth.authenticate()

    os.environ.__setitem__.assert_called_once_with("MLFLOW_TEST_ENV_VAR", "access_token")



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import pytest

from anemoi.utils.compatibility import aliases


def test_aliases() -> None:
    """Test the aliases decorator for function argument aliases."""

    @aliases(a="b", c=["d", "e"])
    def func(a, c):
        return a, c

    assert func(a=1, c=2) == (1, 2)
    assert func(a=1, d=2) == (1, 2)
    assert func(b=1, d=2) == (1, 2)


def test_duplicate_values() -> None:
    """Test the aliases decorator for handling duplicate values."""

    @aliases(a="b", c=["d", "e"])
    def func(a, c):
        return a, c

    with pytest.raises(ValueError):
        func(a=1, b=2)



# (C) Copyright 2024 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import os
import shutil
import sys

import pytest

from anemoi.utils.remote import TransferMethodNotImplementedError
from anemoi.utils.remote import _find_transfer_class
from anemoi.utils.remote import transfer
from anemoi.utils.testing import packages_installed

IN_CI = (os.environ.get("GITHUB_WORKFLOW") is not None) or (os.environ.get("IN_CI_HPC") is not None)

LOCAL = [
    "/absolute/path/to/file",
    "relative/file",
    "/absolute/path/to/dir/",
    "relative/dir/",
    ".",
    "..",
    "./",
    "file",
    "dir/",
    "/dir/",
    "/dir",
    "/file",
]
S3 = ["s3://bucket/key/", "s3://bucket/key"]
SSH = [
    "ssh://hostname:/absolute/file",
    "ssh://hostname:relative/file",
    "ssh://hostname:/absolute/dir/",
    "ssh://hostname:relative/dir/",
]

ROOT_S3_READ = "s3://ml-tests/test-data/anemoi-utils/pytest/transfer"
ROOT_S3_WRITE = f"s3://ml-tmp/anemoi-utils/pytest/transfer/test-{os.getpid()}"

LOCAL_TEST_DATA = os.path.dirname(__file__) + "/test-transfer-data"


@pytest.mark.parametrize("source", LOCAL)
@pytest.mark.parametrize("target", S3)
def test_transfer_find_s3_upload(source: str, target: str) -> None:
    """Test finding the S3 upload transfer class.

    Parameters
    ----------
    source : str
        The source path
    target : str
        The target path
    """
    from anemoi.utils.remote.s3 import S3Upload

    assert _find_transfer_class(source, target) == S3Upload


@pytest.mark.parametrize("source", S3)
@pytest.mark.parametrize("target", LOCAL)
def test_transfer_find_s3_download(source: str, target: str) -> None:
    """Test finding the S3 download transfer class.

    Parameters
    ----------
    source : str
        The source path
    target : str
        The target path
    """
    from anemoi.utils.remote.s3 import S3Download

    assert _find_transfer_class(source, target) == S3Download


@pytest.mark.parametrize("source", LOCAL)
@pytest.mark.parametrize("target", SSH)
def test_transfer_find_ssh_upload(source: str, target: str) -> None:
    """Test finding the SSH upload transfer class.

    Parameters
    ----------
    source : str
        The source path
    target : str
        The target path
    """
    from anemoi.utils.remote.ssh import RsyncUpload

    assert _find_transfer_class(source, target) == RsyncUpload


@pytest.mark.parametrize("source", S3 + SSH)
@pytest.mark.parametrize("target", S3 + SSH)
def test_transfer_find_none(source: str, target: str) -> None:
    """Test that no transfer class is found for unsupported transfers.

    Parameters
    ----------
    source : str
        The source path
    target : str
        The target path
    """
    with pytest.raises(TransferMethodNotImplementedError):
        assert _find_transfer_class(source, target)


@pytest.mark.skipif(IN_CI, reason="Test requires access to S3")
@pytest.mark.skipif(not packages_installed("boto3"), reason="boto3 is not installed")
def test_transfer_zarr_s3_to_local(tmpdir: pytest.TempPathFactory) -> None:
    """Test transferring a Zarr file from S3 to local.

    Parameters
    ----------
    tmpdir : pytest.TempPathFactory
        Temporary directory factory
    """
    source = "s3://ml-datasets/aifs-ea-an-oper-0001-mars-20p0-2000-2000-12h-v0-TESTING2.zarr/"
    tmp = tmpdir.strpath + "/test"

    transfer(source, tmp)
    with pytest.raises(ValueError, match="already exists"):
        transfer(source, tmp)

    transfer(source, tmp, resume=True)
    transfer(source, tmp, overwrite=True)


@pytest.mark.skipif(IN_CI, reason="Test requires access to S3")
@pytest.mark.skipif(not packages_installed("boto3"), reason="boto3 is not installed")
def test_transfer_zarr_local_to_s3(tmpdir: pytest.TempPathFactory) -> None:
    """Test transferring a Zarr file from local to S3.

    Parameters
    ----------
    tmpdir : pytest.TempPathFactory
        Temporary directory factory
    """
    fixture = "s3://ml-datasets/aifs-ea-an-oper-0001-mars-20p0-2000-2000-12h-v0-TESTING2.zarr/"
    source = tmpdir.strpath + "/test"
    target = ROOT_S3_WRITE + "/test.zarr"

    transfer(fixture, source)
    transfer(source, target)

    with pytest.raises(ValueError, match="already exists"):
        transfer(source, target)

    transfer(source, target, resume=True)
    transfer(source, target, overwrite=True)


def _delete_file_or_directory(path: str) -> None:
    """Delete a file or directory.

    Parameters
    ----------
    path : str
        The path to the file or directory
    """
    if os.path.isdir(path):
        shutil.rmtree(path, ignore_errors=True)
    else:
        if os.path.exists(path):
            os.remove(path)


def compare(local1: str, local2: str) -> None:
    """Compare two local files or directories.

    Parameters
    ----------
    local1 : str
        The path to the first file or directory
    local2 : str
        The path to the second file or directory
    """
    if os.path.isdir(local1):
        for root, dirs, files in os.walk(local1):
            for file in files:
                file1 = os.path.join(root, file)
                file2 = file1.replace(local1, local2)
                assert os.path.exists(file2)
                with open(file1, "rb") as f1, open(file2, "rb") as f2:
                    assert f1.read() == f2.read()
    else:
        with open(local1, "rb") as f1, open(local2, "rb") as f2:
            assert f1.read() == f2.read()


@pytest.mark.skipif(IN_CI, reason="Test requires access to S3")
@pytest.mark.skipif(not packages_installed("boto3"), reason="boto3 is not installed")
@pytest.mark.parametrize("path", ["directory/", "file"])
def test_transfer_local_to_s3_to_local(path: str) -> None:
    """Test transferring a file or directory from local to S3 and back to local.

    Parameters
    ----------
    path : str
        The path to the file or directory
    """
    local = LOCAL_TEST_DATA + "/" + path
    remote = ROOT_S3_WRITE + "/" + path
    local2 = LOCAL_TEST_DATA + "-copy-" + path

    transfer(local, remote, overwrite=True)
    transfer(local, remote, resume=True)
    with pytest.raises(ValueError, match="already exists"):
        transfer(local, remote)

    _delete_file_or_directory(local2)
    transfer(remote, local2)
    with pytest.raises(ValueError, match="already exists"):
        transfer(remote, local2)
    transfer(local, remote, overwrite=True)
    transfer(local, remote, resume=True)

    compare(local, local2)

    _delete_file_or_directory(local2)


@pytest.mark.skipif(IN_CI, reason="Test requires ssh access to localhost")
@pytest.mark.skipif(sys.platform == "darwin", reason="Does not work on MacOS")
@pytest.mark.parametrize("path", ["directory", "file"])
@pytest.mark.parametrize("temporary_target", [True, False])
def test_transfer_local_to_ssh(path: str, temporary_target: bool) -> None:
    """Test transferring a file or directory from local to SSH.

    Parameters
    ----------
    path : str
        The path to the file or directory
    temporary_target : bool
        Whether to use a temporary target
    """
    local = LOCAL_TEST_DATA + "/" + path
    remote_path = LOCAL_TEST_DATA + "-as-ssh-" + path
    assert os.path.isabs(remote_path), remote_path

    remote = "ssh://localhost:" + remote_path

    transfer(local, remote, temporary_target=temporary_target)
    transfer(local, remote, temporary_target=temporary_target)

    compare(local, remote_path)

    _delete_file_or_directory(remote_path)


if __name__ == "__main__":
    for name, obj in list(globals().items()):
        if name.startswith("test_") and callable(obj):
            print(f"Running {name}...")
            obj()



# (C) Copyright 2024 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import sys
from typing import Any
from typing import TextIO


def print_request(verb: str, request: dict[str, Any], file: TextIO = sys.stdout) -> None:
    """Prints a formatted request.

    Parameters
    ----------
    verb : str
        A mars verb
    request : Dict[str, Any]
        The request parameters.
    file : TextIO, optional
        The file to which the request is printed, by default sys.stdout.
    """
    r = [verb]
    for k, v in request.items():
        if not isinstance(v, (list, tuple, set)):
            v = [v]
        v = [str(_) for _ in v]
        v = "/".join(v)
        r.append(f"{k}={v}")

    r = ",\n   ".join(r)
    print(r, file=file)
    print(file=file)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from __future__ import annotations

import contextlib
import json
import logging
import os
import threading
import warnings
from typing import Any

import yaml

try:
    import tomllib  # Only available since 3.11
except ImportError:
    import tomli as tomllib


LOG = logging.getLogger(__name__)


class DotDict(dict):
    """A dictionary that allows access to its keys as attributes.

    >>> d = DotDict({"a": 1, "b": {"c": 2}})
    >>> d.a
    1
    >>> d.b.c
    2
    >>> d.b = 3
    >>> d.b
    3

    The class is recursive, so nested dictionaries are also DotDicts.

    The DotDict class has the same constructor as the dict class.

    >>> d = DotDict(a=1, b=2)
    """

    def __init__(self, *args, **kwargs):
        """Initialize a DotDict instance.

        Parameters
        ----------
        *args : tuple
            Positional arguments for the dict constructor.
        **kwargs : dict
            Keyword arguments for the dict constructor.
        """
        super().__init__(*args, **kwargs)

        for k, v in self.items():
            self[k] = self.convert_to_nested_dot_dict(v)

    @staticmethod
    def convert_to_nested_dot_dict(value):
        if isinstance(value, dict) or is_omegaconf_dict(value):
            return DotDict(value)

        if isinstance(value, list) or is_omegaconf_list(value):
            return [DotDict(i) if isinstance(i, dict) or is_omegaconf_dict(i) else i for i in value]

        if isinstance(value, tuple):
            return [DotDict(i) if isinstance(i, dict) or is_omegaconf_dict(i) else i for i in value]

        return value

    @classmethod
    def from_file(cls, path: str) -> DotDict:
        """Create a DotDict from a file.

        Parameters
        ----------
        path : str
            The path to the file.

        Returns
        -------
        DotDict
            The created DotDict.
        """
        _, ext = os.path.splitext(path)
        if ext == ".yaml" or ext == ".yml":
            return cls.from_yaml_file(path)
        elif ext == ".json":
            return cls.from_json_file(path)
        elif ext == ".toml":
            return cls.from_toml_file(path)
        else:
            raise ValueError(f"Unknown file extension {ext}")

    @classmethod
    def from_yaml_file(cls, path: str) -> DotDict:
        """Create a DotDict from a YAML file.

        Parameters
        ----------
        path : str
            The path to the YAML file.

        Returns
        -------
        DotDict
            The created DotDict.
        """
        with open(path) as file:
            data = yaml.safe_load(file)

        return cls(data)

    @classmethod
    def from_json_file(cls, path: str) -> DotDict:
        """Create a DotDict from a JSON file.

        Parameters
        ----------
        path : str
            The path to the JSON file.

        Returns
        -------
        DotDict
            The created DotDict.
        """
        with open(path) as file:
            data = json.load(file)

        return cls(data)

    @classmethod
    def from_toml_file(cls, path: str) -> DotDict:
        """Create a DotDict from a TOML file.

        Parameters
        ----------
        path : str
            The path to the TOML file.

        Returns
        -------
        DotDict
            The created DotDict.
        """
        with open(path) as file:
            data = tomllib.load(file)
        return cls(data)

    def __getattr__(self, attr: str) -> Any:
        """Get an attribute.

        Parameters
        ----------
        attr : str
            The attribute name.

        Returns
        -------
        Any
            The attribute value.
        """
        try:
            return self[attr]
        except KeyError:
            raise AttributeError(attr)

    def __setattr__(self, attr: str, value: Any) -> None:
        """Set an attribute.

        Parameters
        ----------
        attr : str
            The attribute name.
        value : Any
            The attribute value.
        """

        self.warn_on_mutation(attr)
        value = self.convert_to_nested_dot_dict(value)
        super().__setitem__(attr, value)

    def __setitem__(self, key: str, value: Any) -> None:
        """Set an item in the dictionary.

        Parameters
        ----------
        key : str
            The key to set.
        value : Any
            The value to set.
        """
        self.warn_on_mutation(key)
        value = self.convert_to_nested_dot_dict(value)
        super().__setitem__(key, value)

    def warn_on_mutation(self, key):
        warnings.warn("Mofifying and instance of DotDict(). This class is intended to be immutable.")

    def __repr__(self) -> str:
        """Return a string representation of the DotDict.

        Returns
        -------
        str
            The string representation.
        """
        return f"DotDict({super().__repr__()})"


def is_omegaconf_dict(value: Any) -> bool:
    """Check if a value is an OmegaConf DictConfig.

    Parameters
    ----------
    value : Any
        The value to check.

    Returns
    -------
    bool
        True if the value is a DictConfig, False otherwise.
    """
    try:
        from omegaconf import DictConfig

        return isinstance(value, DictConfig)
    except ImportError:
        return False


def is_omegaconf_list(value: Any) -> bool:
    """Check if a value is an OmegaConf ListConfig.

    Parameters
    ----------
    value : Any
        The value to check.

    Returns
    -------
    bool
        True if the value is a ListConfig, False otherwise.
    """
    try:
        from omegaconf import ListConfig

        return isinstance(value, ListConfig)
    except ImportError:
        return False


CONFIG = {}
CHECKED = {}
CONFIG_LOCK = threading.RLock()
QUIET = False
CONFIG_PATCH = None


def _find(config: dict | list, what: str, result: list = None) -> list:
    """Find all occurrences of a key in a nested dictionary or list.

    Parameters
    ----------
    config : dict or list
        The configuration to search.
    what : str
        The key to search for.
    result : list, optional
        The list to store results, by default None.

    Returns
    -------
    list
        The list of found values.
    """
    if result is None:
        result = []

    if isinstance(config, list):
        for i in config:
            _find(i, what, result)
        return result

    if isinstance(config, dict):
        if what in config:
            result.append(config[what])

        for k, v in config.items():
            _find(v, what, result)

    return result


def _merge_dicts(a: dict, b: dict) -> None:
    """Merge two dictionaries recursively.

    Parameters
    ----------
    a : dict
        The first dictionary.
    b : dict
        The second dictionary.
    """
    for k, v in b.items():
        if k in a and isinstance(a[k], dict) and isinstance(v, dict):
            _merge_dicts(a[k], v)
        else:
            a[k] = v


def _set_defaults(a: dict, b: dict) -> None:
    """Set default values in a dictionary.

    Parameters
    ----------
    a : dict
        The dictionary to set defaults in.
    b : dict
        The dictionary with default values.
    """
    for k, v in b.items():
        if k in a and isinstance(a[k], dict) and isinstance(v, dict):
            _set_defaults(a[k], v)
        else:
            a.setdefault(k, v)


def config_path(name: str = "settings.toml") -> str:
    """Get the path to a configuration file.

    Parameters
    ----------
    name : str, optional
        The name of the configuration file, by default "settings.toml".

    Returns
    -------
    str
        The path to the configuration file.
    """
    global QUIET

    if name.startswith("/") or name.startswith("."):
        return name

    if name.startswith("~"):
        return os.path.expanduser(name)

    full = os.path.join(os.path.expanduser("~"), ".config", "anemoi", name)
    os.makedirs(os.path.dirname(full), exist_ok=True)

    if name == "settings.toml":
        old = os.path.join(os.path.expanduser("~"), ".anemoi.toml")
        if not os.path.exists(full) and os.path.exists(old):
            if not QUIET:
                LOG.warning(
                    "Configuration file found at ~/.anemoi.toml. Please move it to ~/.config/anemoi/settings.toml"
                )
                QUIET = True
            return old
        else:
            if os.path.exists(old):
                if not QUIET:
                    LOG.warning(
                        "Configuration file found at ~/.anemoi.toml and ~/.config/anemoi/settings.toml, ignoring the former"
                    )
                    QUIET = True

    return full


def load_any_dict_format(path: str) -> dict:
    """Load a configuration file in any supported format: JSON, YAML and TOML.

    Parameters
    ----------
    path : str
        The path to the configuration file.

    Returns
    -------
    dict
        The decoded configuration file.
    """

    try:
        if path.endswith(".json"):
            with open(path, "rb") as f:
                return json.load(f)

        if path.endswith(".yaml") or path.endswith(".yml"):
            with open(path, "rb") as f:
                return yaml.safe_load(f)

        if path.endswith(".toml"):
            with open(path, "rb") as f:
                return tomllib.load(f)

        if path == "-":
            import sys

            config = sys.stdin.read()

            parsers = [(yaml.safe_load, "yaml"), (json.loads, "json"), (tomllib.loads, "toml")]

            for parser, parser_type in parsers:
                try:
                    LOG.debug(f"Trying {parser_type} parser for stdin")
                    return parser(config)
                except Exception:
                    pass

            raise ValueError("Failed to parse configuration from stdin")

    except (json.JSONDecodeError, yaml.YAMLError, tomllib.TOMLDecodeError) as e:
        LOG.warning(f"Failed to parse config file {path}", exc_info=e)
        raise ValueError(f"Failed to parse config file {path} [{e}]")

    return open(path).read()


def _load_config(
    name: str = "settings.toml",
    secrets: str | list[str] | None = None,
    defaults: str | dict | None = None,
) -> DotDict:
    """Load a configuration file.

    Parameters
    ----------
    name : str, optional
        The name of the configuration file, by default "settings.toml".
    secrets : str or list, optional
        The name of the secrets file, by default None.
    defaults : str or dict, optional
        The name of the defaults file, by default None.

    Returns
    -------
    DotDict
        The loaded configuration.
    """
    key = json.dumps((name, secrets, defaults), sort_keys=True, default=str)
    if key in CONFIG:
        return CONFIG[key]

    path = config_path(name)
    if os.path.exists(path):
        config = load_any_dict_format(path)
    else:
        config = {}

    if defaults is not None:
        if isinstance(defaults, str):
            defaults = load_raw_config(defaults)
        _set_defaults(config, defaults)

    if secrets is not None:
        if isinstance(secrets, str):
            secrets = [secrets]

        base, ext = os.path.splitext(path)
        secret_name = base + ".secrets" + ext

        found = set()
        for secret in secrets:
            if _find(config, secret):
                found.add(secret)

        if found:
            check_config_mode(name, secret_name, found)

        check_config_mode(secret_name, None)
        secret_config = _load_config(secret_name)
        _merge_dicts(config, secret_config)

    for env, value in os.environ.items():

        if not env.startswith("ANEMOI_CONFIG_"):
            continue
        rest = env[len("ANEMOI_CONFIG_") :]

        package = rest.split("_")[0]
        sub = rest[len(package) + 1 :]

        package = package.lower()
        sub = sub.lower()

        LOG.info(f"Using environment variable {env} to override the anemoi config key '{package}.{sub}'")

        if package not in config:
            config[package] = {}
        config[package][sub] = value

    CONFIG[key] = DotDict(config)
    return CONFIG[key]


def _save_config(name: str, data: Any) -> None:
    """Save a configuration file.

    Parameters
    ----------
    name : str
        The name of the configuration file.
    data : Any
        The data to save.
    """
    CONFIG.pop(name, None)

    conf = config_path(name)

    if conf.endswith(".json"):
        with open(conf, "w") as f:
            json.dump(data, f, indent=4)
        return

    if conf.endswith(".yaml") or conf.endswith(".yml"):
        with open(conf, "w") as f:
            yaml.dump(data, f)
        return

    if conf.endswith(".toml"):
        raise NotImplementedError("Saving to TOML is not implemented yet")

    with open(conf, "w") as f:
        f.write(data)


def save_config(name: str, data: Any) -> None:
    """Save a configuration file.

    Parameters
    ----------
    name : str
        The name of the configuration file to save.

    data : Any
        The data to save.
    """
    with CONFIG_LOCK:
        _save_config(name, data)


def load_config(
    name: str = "settings.toml",
    secrets: str | list[str] | None = None,
    defaults: str | dict | None = None,
) -> DotDict | str:
    """Read a configuration file.

    Parameters
    ----------
    name : str, optional
        The name of the config file to read, by default "settings.toml"
    secrets : str or list, optional
        The name of the secrets file, by default None
    defaults : str or dict, optional
        The name of the defaults file, by default None

    Returns
    -------
    DotDict or str
        Return DotDict if it is a dictionary, otherwise the raw data
    """

    with CONFIG_LOCK:
        config = _load_config(name, secrets, defaults)
        if CONFIG_PATCH is not None:
            config = CONFIG_PATCH(config)
        return config


def load_raw_config(name: str, default: Any = None) -> DotDict | str:
    """Load a raw configuration file.

    Parameters
    ----------
    name : str
        The name of the configuration file.
    default : Any, optional
        The default value if the file does not exist, by default None.

    Returns
    -------
    DotDict or str
        The loaded configuration or the default value.
    """
    path = config_path(name)
    if os.path.exists(path):
        return load_any_dict_format(path)

    return default


def check_config_mode(name: str = "settings.toml", secrets_name: str = None, secrets: list[str] = None) -> None:
    """Check that a configuration file is secure.

    Parameters
    ----------
    name : str, optional
        The name of the configuration file, by default "settings.toml"
    secrets_name : str, optional
        The name of the secrets file, by default None
    secrets : list, optional
        The list of secrets to check, by default None

    Raises
    ------
    SystemError
        If the configuration file is not secure.
    """
    with CONFIG_LOCK:
        if name in CHECKED:
            return

        conf = config_path(name)
        if not os.path.exists(conf):
            return
        mode = os.stat(conf).st_mode
        if mode & 0o777 != 0o600:
            if secrets_name:
                secret_path = config_path(secrets_name)
                raise SystemError(
                    f"Configuration file {conf} should not hold entries {secrets}.\n"
                    f"Please move them to {secret_path}."
                )
            raise SystemError(f"Configuration file {conf} is not secure.\n" f"Please run `chmod 600 {conf}`.")

        CHECKED[name] = True


def find(metadata: dict | list, what: str, result: list = None, *, select: callable = None) -> list:
    """Find all occurrences of a key in a nested dictionary or list with an optional selector.

    Parameters
    ----------
    metadata : dict or list
        The metadata to search.
    what : str
        The key to search for.
    result : list, optional
        The list to store results, by default None.
    select : callable, optional
        A function to filter the results, by default None.

    Returns
    -------
    list
        The list of found values.
    """
    if result is None:
        result = []

    if isinstance(metadata, list):
        for i in metadata:
            find(i, what, result)
        return result

    if isinstance(metadata, dict):
        if what in metadata:
            if select is None or select(metadata[what]):
                result.append(metadata[what])

        for k, v in metadata.items():
            find(v, what, result)

    return result


def merge_configs(*configs: dict) -> dict:
    """Merge multiple configuration dictionaries.

    Parameters
    ----------
    *configs : dict
        The configuration dictionaries to merge.

    Returns
    -------
    dict
        The merged configuration dictionary.
    """
    result = {}
    for config in configs:
        _merge_dicts(result, config)

    return result


@contextlib.contextmanager
def temporary_config(tmp: dict) -> None:

    global CONFIG_PATCH

    def patch_config(config: dict) -> dict:
        return merge_configs(config, tmp)

    with CONFIG_LOCK:

        CONFIG_PATCH = patch_config

        try:
            yield
        finally:
            CONFIG_PATCH = None



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import importlib
import logging
import os
import sys
import warnings
from collections.abc import Callable
from functools import cached_property
from typing import Any
from typing import Generic
from typing import Optional
from typing import TypeVar

import entrypoints

LOG = logging.getLogger(__name__)

DEBUG_ANEMOI_REGISTRY = int(os.environ.get("DEBUG_ANEMOI_REGISTRY", "0"))


class Wrapper:
    """A wrapper for the registry.

    Parameters
    ----------
    name : str
        The name of the wrapper.
    registry : Registry
        The registry to wrap.
    """

    def __init__(self, name: str, registry: "Registry"):
        self.name = name
        self.registry = registry

    def __call__(self, factory: Callable) -> Callable:
        """Register a factory with the registry.

        Parameters
        ----------
        factory : Callable
            The factory to register.

        Returns
        -------
        Callable
            The registered factory.
        """
        self.registry.register(self.name, factory)
        return factory


class Error:
    """An error class. Used in place of a plugin that failed to load.

    Parameters
    ----------
    error : Exception
        The error.
    """

    def __init__(self, error: Exception):
        self.error = error

    def __call__(self, *args, **kwargs):
        raise self.error


_BY_KIND = {}

T = TypeVar("T")


class Registry(Generic[T]):
    """A registry of factories.

    Parameters
    ----------
    package : str
        The package name.
    key : str, optional
        The key to use for the registry, by default "_type".
    api_version : str, optional
        The API version, by default '1.0.0'.
    """

    def __init__(self, package: str, key: str = "_type", api_version: str = "1.0.0"):
        self.package = package
        self.__registered = {}
        self._sources = {}
        self.kind = package.split(".")[-1]
        self.key = key
        self.api_version = api_version
        _BY_KIND[self.kind] = self

    @classmethod
    def lookup_kind(cls, kind: str) -> Optional["Registry"]:
        """Lookup a registry by kind.

        Parameters
        ----------
        kind : str
            The kind of the registry.

        Returns
        -------
        Registry, optional
            The registry if found, otherwise None.
        """
        return _BY_KIND.get(kind)

    def register(self, name: str, factory: Callable | None = None, source: Any | None = None) -> Wrapper | None:
        """Register a factory with the registry.

        Parameters
        ----------
        name : str
            The name of the factory.
        factory : Callable, optional
            The factory to register, by default None.
        source : Any, optional
            The source of the factory, by default None.

        Returns
        -------
        Wrapper, optional
            A wrapper if the factory is None, otherwise None.
        """

        name = name.replace("_", "-")

        if factory is None:
            # This happens when the @register decorator is used
            return Wrapper(name, self)

        if source is None:
            source = getattr(factory, "_source") if hasattr(factory, "_source") else factory

        if name in self.__registered:
            warnings.warn(f"Factory '{name}' is already registered in {self.package}")
            warnings.warn(f"Existing: {self._sources[name]}")
            warnings.warn(f"New: {source}")

        self.__registered[name] = factory
        self._sources[name] = source

    def _load(self, file: str) -> None:
        """Load a module from a file.

        Parameters
        ----------
        file : str
            The file to load.
        """

        name, _ = os.path.splitext(file)

        try:
            importlib.import_module(f".{name}", package=self.package)
        except Exception as e:
            if DEBUG_ANEMOI_REGISTRY:
                raise

            name = name.replace("_", "-")
            self.__registered[name] = Error(e)

    def is_registered(self, name: str) -> bool:
        """Check if a factory is registered.

        Parameters
        ----------
        name : str
            The name of the factory.

        Returns
        -------
        bool
            Whether the factory is registered.
        """

        name = name.replace("_", "-")

        ok = name in self.factories
        if not ok:
            LOG.error(f"Cannot find '{name}' in {self.package}")
            for e in self.factories:
                LOG.info(f"Registered: {e} ({self._sources.get(e)})")
        return ok

    def lookup(self, name: str, *, return_none: bool = False) -> Callable | None:
        """Lookup a factory by name.

        Parameters
        ----------
        name : str
            The name of the factory.
        return_none : bool, optional
            Whether to return None if the factory is not found, by default False.

        Returns
        -------
        Callable, optional
            The factory if found, otherwise None.
        """

        name = name.replace("_", "-")

        if return_none:
            return self.factories.get(name)

        factory = self.factories.get(name)
        if factory is None:

            LOG.error(f"Cannot find '{name}' in {self.package}")
            for e in self.factories:
                LOG.info(f"Registered: {e} ({self._sources.get(e)})")

            raise ValueError(f"Cannot find '{name}' in {self.package}")

        return factory

    @cached_property
    def factories(self) -> dict[str, Callable]:

        directory = sys.modules[self.package].__path__[0]

        for file in os.listdir(directory):

            if file[0] == ".":
                continue

            if file == "__init__.py":
                continue

            full = os.path.join(directory, file)
            if os.path.isdir(full):
                if os.path.exists(os.path.join(full, "__init__.py")):
                    self._load(file)
                continue

            if file.endswith(".py"):
                self._load(file)

        bits = self.package.split(".")
        # We assume a name like anemoi.datasets.create.sources, with kind = sources
        assert bits[-1] == self.kind, (self.package, self.kind)
        assert len(bits) > 1, self.package

        groups = []
        middle = bits[1:-1]
        while True:
            group = ".".join([bits[0], *middle, bits[-1]])
            groups.append(group)
            if len(middle) == 0:
                break
            middle.pop()

        groups.reverse()

        LOG.debug("Loading plugins from %s", groups)

        for entrypoint_group in groups:
            for entry_point in entrypoints.get_group_all(entrypoint_group):
                source = entry_point.distro
                try:
                    self.register(entry_point.name, entry_point.load(), source=source)
                except Exception as e:
                    if DEBUG_ANEMOI_REGISTRY:
                        raise
                    self.register(entry_point.name, Error(e), source=source)

        return self.__registered

    @property
    def registered(self) -> list[str]:
        """Get the registered factories."""

        return sorted(self.factories.keys())

    def create(self, name: str, *args: Any, **kwargs: Any) -> T:
        """Create an instance using a factory.

        Parameters
        ----------
        name : str
            The name of the factory.
        *args : Any
            Positional arguments for the factory.
        **kwargs : Any
            Keyword arguments for the factory.

        Returns
        -------
        Any
            The created instance.
        """

        name = name.replace("_", "-")

        factory = self.lookup(name)
        return factory(*args, **kwargs)

    def from_config(self, config: str | dict[str, Any], *args: Any, **kwargs: Any) -> T:
        """Create an instance from a configuration.

        Parameters
        ----------
        config : str or dict
            The configuration.
        *args : Any
            Positional arguments for the factory.
        **kwargs : Any
            Keyword arguments for the factory.

        Returns
        -------
        Any
            The created instance.
        """
        if isinstance(config, str):
            config = {config: {}}

        if not isinstance(config, dict):
            raise ValueError(f"Invalid config: {config}")

        if self.key in config:
            config = config.copy()
            key = config.pop(self.key)
            return self.create(key, *args, **config, **kwargs)

        if len(config) == 1:
            key = list(config.keys())[0]
            value = config[key]

            if isinstance(value, dict):
                return self.create(key, *args, **value, **kwargs)

            if isinstance(value, list):
                return self.create(key, *args, *value, **kwargs)

            return self.create(key, *args, value, **kwargs)

        raise ValueError(
            f"Entry '{config}' must either be a string, a dictionary with a single entry, or a dictionary with a '{self.key}' key"
        )



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from anemoi.utils import provenance


def test_gather() -> None:
    """Test success of gather_provenance_info."""
    provenance.gather_provenance_info()
